import subprocess
import sys

from typing import Optional
from typing import List

from starlette.requests import Request
from starlette.responses import StreamingResponse

from spinta.formats.ascii.components import Ascii
from spinta.components import Context, Action, UrlParams, Model
from spinta import commands
from spinta.utils.response import aiter


@commands.render.register(Context, Request, Model, Ascii)
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: Ascii,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    return _render(context, model, fmt, action, params, data, status_code, headers)


def _render(
    context: Context,
    model: Model,
    fmt: Ascii,
    action: Action,
    params: UrlParams,
    data,
    status_code,
    headers,
):
    # Format params ar given in RQL query `?format(width(1),colwidth(1))`.
    width = params.formatparams.get('width')
    colwidth = params.formatparams.get('colwidth')

    if width is None and colwidth is None and sys.stdin.isatty():
        _, width = map(int, subprocess.run(['stty', 'size'], capture_output=True).stdout.split())
    elif width is None and colwidth is None:
        colwidth = 42

    return StreamingResponse(
        aiter(fmt(context, model, action, params, data, width, colwidth)),
        status_code=status_code,
        media_type=fmt.content_type,
        headers=headers,
    )



import base64
import hashlib
import hmac
import math
import os
import random


def _bencode(value: bytes) -> str:
    return base64.urlsafe_b64encode(value).decode().rstrip('=')


def _bdecode(value: str) -> bytes:
    value = value.ljust(math.ceil(len(value) / 4) * 4, '=')
    return base64.urlsafe_b64decode(value)


def crypt(password, iterations=None, salt=None):
    if iterations is None:
        iterations = random.randint(100_000, 999_999)
    if salt is None:
        salt = os.urandom(16)
    elif isinstance(salt, str):
        salt = salt.encode()
    if isinstance(password, str):
        password = password.encode()
    hash = hashlib.pbkdf2_hmac('sha256', password, salt, iterations)
    hash = _bencode(hash)
    salt = _bencode(salt)
    return f'pbkdf2$sha256${iterations}${salt}${hash}'


def verify(password, hash):
    if isinstance(password, str):
        password = password.encode()

    hasher, *args = hash.split('$')
    if hasher == 'pbkdf2':
        hash_name, iterations, salt, hash = args
        iterations = int(iterations)
        salt = _bdecode(salt)
        given = _bdecode(hash)
        repro = hashlib.pbkdf2_hmac(hash_name, password, salt, iterations)
    else:
        raise Exception(f"Unknown hasher {hasher!r}.")

    return hmac.compare_digest(given, repro)


def gensecret(size=32):
    n = math.floor(math.log(64, 256) * size) + 1
    return _bencode(os.urandom(n))[:size]



from contextlib import contextmanager
from pathlib import Path

from spinta.utils.refs import get_ref_id
from spinta.components import Context


def _get_cache_path(base: Path, key: str) -> Path:
    return base / key[:2] / key[2:4] / key[4:]


class Cache:

    def __init__(self, path: Path):
        self.path = path

    def get(self, key: object):
        key = get_ref_id(key)
        path = _get_cache_path(self.path, key)
        if path.exists():
            return path

    @contextmanager
    def set(self, key: object, *, text: bool = False):
        key = get_ref_id(key)
        path = _get_cache_path(self.path, key)
        path.parent.mkdir(parents=True, exist_ok=True)
        mode = 'wt' if text else 'wb'
        encoding = 'utf-8' if text else None
        with path.open(mode, encoding=encoding) as f:
            yield f


def fetch(context: Context, url: str, *, text: bool = False) -> Path:
    cache = context.get('cache')
    path = cache.get(url)
    if path is not None:
        return path

    requests = context.get('requests')
    with requests.get(url, stream=True) as r:
        with cache.set(url, text=text) as f:
            chunks = r.iter_content(chunk_size=8192, decode_unicode=text)
            for chunk in filter(None, chunks):
                f.write(chunk)

    return cache.get(url)



import dataclasses
import datetime
from decimal import Decimal
from typing import Any
from typing import Dict
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import TypeVar

import pkg_resources as pres
from itertools import chain
from itertools import count
from starlette.requests import Request
from starlette.templating import Jinja2Templates

from spinta import commands
from spinta.backends.components import SelectTree
from spinta.backends.helpers import get_model_reserved_props
from spinta.backends.helpers import get_ns_reserved_props
from spinta.backends.helpers import select_model_props
from spinta.components import Action
from spinta.components import Context
from spinta.components import Model
from spinta.components import Namespace
from spinta.components import UrlParams
from spinta.formats.components import Format
from spinta.formats.helpers import get_model_tabular_header
from spinta.formats.html.components import Cell
from spinta.formats.html.components import Color
from spinta.formats.html.components import ComplexCell
from spinta.formats.html.components import Html
from spinta.formats.html.helpers import get_model_link
from spinta.formats.html.helpers import get_output_formats
from spinta.formats.html.helpers import get_template_context
from spinta.formats.html.helpers import short_id
from spinta.types.datatype import Array
from spinta.types.datatype import DataType
from spinta.types.datatype import File
from spinta.types.datatype import Object
from spinta.types.datatype import Ref
from spinta.types.datatype import String
from spinta.types.datatype import Date
from spinta.types.datatype import Time
from spinta.types.datatype import DateTime
from spinta.types.datatype import Number
from spinta.types.datatype import Binary
from spinta.types.datatype import JSON
from spinta.utils.nestedstruct import flatten
from spinta.utils.schema import NotAvailable


def _get_model_reserved_props(action: Action) -> List[str]:
    if action in (Action.GETALL, Action.SEARCH):
        return ['_id']
    else:
        return get_model_reserved_props(action)


def _render_check(request: Request, data: Dict[str, Any] = None):
    if data:
        if 'errors' in data:
            result = {
                'message': "Yra klaidų",
                'errors': [
                    err['message']
                    for err in data['errors']
                ],
            }
        else:
            result = {
                'message': "Klaidų nerasta.",
                'errors': None,
            }
    else:
        result = data

    templates = Jinja2Templates(directory=pres.resource_filename('spinta', 'templates'))
    return templates.TemplateResponse('form.html', {
        'request': request,
        'title': "Duomenų struktūros aprašo tikrinimas",
        'description': (
            "Ši priemonė leidžia patikrinti ar "
            "<a href=\"https://atviriduomenys.readthedocs.io/dsa/index.html\">"
            "duomenų struktūros apraše</a> nėra klaidų."
        ),
        'name': 'check',
        'fields': [
            {
                'label': "Duomenų struktūros aprašas",
                'help': "Pateikite duomenų struktūros aprašo failą.",
                'input': '<input type="file" name="manifest" accept=".csv">'
            },
        ],
        'submit': "Tikrinti",
        'result': result,
    })


@commands.render.register(Context, Request, Namespace, Html)
def render(
    context: Context,
    request: Request,
    ns: Namespace,
    fmt: Html,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    if action == Action.CHECK:
        return _render_check(request, data)
    else:
        return _render_model(context, request, ns, action, params, data, headers)


@commands.render.register(Context, Request, Model, Html)
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: Html,
    *,
    action: Action,
    params: UrlParams,
    data: Iterator[ComplexCell],
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    if action == Action.GETONE:
        data = [data]
    return _render_model(context, request, model, action, params, data, headers)


T = TypeVar('T')


class _LimitIter(Generic[T]):
    _iterator: Iterator[T]
    limit: Optional[int]
    exhausted: bool = False

    def __init__(self, limit: Optional[int], it: Iterable[T]) -> None:
        self._iterator = iter(it)
        self._counter = count(1)
        self.limit = limit

    def __iter__(self) -> Iterator[T]:
        return self

    def __next__(self) -> T:
        try:
            value = next(self._iterator)
        except StopIteration:
            self.exhausted = True
            raise
        if self.limit is not None and next(self._counter) > self.limit:
            raise StopIteration
        return value


def _is_empty(rows: Iterator[T]) -> Iterator[T]:
    try:
        row = next(rows)
    except StopIteration:
        empty = True
    else:
        empty = False
        rows = chain([row], rows)
    return rows, empty


def _iter_values(
    header: List[str],
    rows: Iterator[Dict[str, T]]
) -> Iterator[List[T]]:
    na = Cell('', color=Color.null)
    for row in rows:
        yield [row.get(h, na) for h in header]


def _get_model_tabular_header(
    context: Context,
    model: Model,
    action: Action,
    params: UrlParams,
) -> List[str]:
    if model.name == '_ns':
        reserved = get_ns_reserved_props(action)
    else:
        reserved = _get_model_reserved_props(action)
    return get_model_tabular_header(
        context,
        model,
        action,
        params,
        reserved=reserved,
    )


def _build_template_context(
    context: Context,
    model: Model,
    action: Action,
    params: UrlParams,
    rows: Iterable[Dict[str, Cell]],
):
    rows = flatten(rows)

    rows, empty = _is_empty(rows)
    header = _get_model_tabular_header(context, model, action, params)
    rows = _iter_values(header, rows)

    if model.name.startswith('_'):
        rows = _LimitIter(None, rows)
    else:
        rows = _LimitIter(params.limit_enforced_to, rows)

    return {
        'header': header,
        'empty': empty,
        'data': rows,
        'params': params,
    }


def _render_model(
    context: Context,
    request: Request,
    model: Model,
    action: Action,
    params: UrlParams,
    rows: Iterable[Dict[str, Cell]],
    http_headers,
):
    ctx = _build_template_context(
        context,
        model,
        action,
        params,
        rows,
    )
    ctx.update(get_template_context(context, model, params))
    ctx['request'] = request
    ctx['formats'] = get_output_formats(params)

    # Pass function references
    ctx['zip'] = zip

    # Preserve response data for tests.
    if request.url.hostname == 'testserver':
        ctx['data'] = list(ctx['data'])

    templates = Jinja2Templates(
        directory=pres.resource_filename('spinta', 'templates')
    )
    return templates.TemplateResponse('data.html', ctx, headers=http_headers)


@dataclasses.dataclass
class _NamespaceName:
    name: str

    def render(self):
        name = self.name.replace('/:ns', '/').rstrip('/').split('/')[-1]
        return f"📁 {name}/"


@dataclasses.dataclass
class _ModelName:
    name: str

    def render(self):
        name = self.name.split('/')[-1]
        return f"📄 {name}"


@commands.prepare_data_for_response.register(Context, Model, Html, dict)
def prepare_data_for_response(
    context: Context,
    model: Model,
    fmt: Html,
    value: dict,
    *,
    action: Action,
    select: SelectTree,
    prop_names: List[str],
) -> dict:
    if model.name == '_ns':
        value = value.copy()
        if value['_type'] == 'ns':
            value['name'] = _NamespaceName(value['name'])
        else:
            value['name'] = _ModelName(value['name'])

    reserved = _get_model_reserved_props(action)

    data = {
        prop.name: commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=value,
            action=action,
            select=sel,
        )
        for prop, val, sel in select_model_props(
            model,
            prop_names,
            value,
            select,
            reserved,
        )
    }

    return data


@commands.prepare_dtype_for_response.register(Context, Html, String, _NamespaceName)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: String,
    value: _NamespaceName,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return Cell(value.render(), link=get_model_link(value))


@commands.prepare_dtype_for_response.register(Context, Html, String, _ModelName)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: String,
    value: _ModelName,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return Cell(value.render(), link=get_model_link(value))


@commands.prepare_dtype_for_response.register(Context, Html, DataType, type(None))
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: DataType,
    value: None,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return Cell('', color=Color.null)


@commands.prepare_dtype_for_response.register(Context, Html, DataType, object)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: DataType,
    value: Any,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    if dtype.prop.name == '_id':
        return Cell(short_id(value), link=get_model_link(
            dtype.prop.model,
            pk=value,
        ))
    return Cell(value)


@commands.prepare_dtype_for_response.register(Context, Html, Date, datetime.date)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: Date,
    value: datetime.date,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return Cell(value.isoformat())


@commands.prepare_dtype_for_response.register(Context, Html, Time, datetime.time)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: Time,
    value: datetime.time,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return Cell(value.isoformat())


@commands.prepare_dtype_for_response.register(Context, Html, DateTime, datetime.datetime)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: DateTime,
    value: datetime.datetime,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return Cell(value.isoformat())


@commands.prepare_dtype_for_response.register(Context, Html, Number, Decimal)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: Number,
    value: Decimal,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return Cell(value)


@commands.prepare_dtype_for_response.register(Context, Html, Binary, bytes)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: Binary,
    value: bytes,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return Cell(value)


@commands.prepare_dtype_for_response.register(Context, Html, JSON, (object, NotAvailable, type(None)))
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: JSON,
    value: Any,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return Cell(value)


@commands.prepare_dtype_for_response.register(Context, Html, Array, tuple)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: Array,
    value: tuple,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    super_ = commands.prepare_dtype_for_response[Context, Format, Array, tuple]
    return super_(context, fmt, dtype, value, data=data, action=action, select=select)


@commands.prepare_dtype_for_response.register(Context, Html, Array, type(None))
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: Array,
    value: None,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    super_ = commands.prepare_dtype_for_response[Context, Format, Array, type(None)]
    return super_(context, fmt, dtype, value, data=data, action=action, select=select)


@commands.prepare_dtype_for_response.register(Context, Html, Ref, (dict, str, type(None)))
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: File,
    value: NotAvailable,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    super_ = commands.prepare_dtype_for_response[Context, Format, Ref, dict]
    return super_(context, fmt, dtype, value, data=data, action=action, select=select)


@commands.prepare_dtype_for_response.register(Context, Html, File, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: File,
    value: NotAvailable,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    super_ = commands.prepare_dtype_for_response[Context, Format, File, NotAvailable]
    return super_(context, fmt, dtype, value, data=data, action=action, select=select)


@commands.prepare_dtype_for_response.register(Context, Html, File, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: File,
    value: Dict[str, Any],
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    if value['_id'] is None:
        _id = Cell('', color=Color.null)
    else:
        _id = Cell(value['_id'], link=get_model_link(
            dtype.prop.model,
            pk=data['_id'],
            prop=dtype.prop.name,
        ))
    if value['_content_type'] is None:
        _content_type = Cell('', color=Color.null)
    else:
        _content_type = Cell(value['_content_type'])
    return {
        '_id': _id,
        '_content_type': _content_type,
    }


@commands.prepare_dtype_for_response.register(Context, Html, Object, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: Object,
    value: Dict[str, Any],
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return commands.prepare_dtype_for_response[Context, Format, Object, dict](
        context,
        fmt,
        dtype,
        value,
        data=data,
        action=action,
        select=select,
    )


@commands.prepare_dtype_for_response.register(Context, Html, Array, list)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: Array,
    value: List[Any],
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return commands.prepare_dtype_for_response[Context, Format, Array, list](
        context,
        fmt,
        dtype,
        value,
        data=data,
        action=action,
        select=select,
    )



from typing import List

import itertools


def build_path_tree(*paths):
    tree = {}
    for path in itertools.chain(*paths):
        add_path_to_tree(tree, path)
    return tree


def add_path_to_tree(tree: dict, path: str):
    parent: List[str] = []
    for part in path.split('/'):
        name = '/'.join(parent)
        parent.append(part)

        if name not in tree:
            tree[name] = [part]
        elif part not in tree[name]:
            tree[name].append(part)
            tree[name].sort()

    name = '/'.join(parent)
    if name not in tree:
        tree[name] = []



from spinta import commands
from spinta.components import Context, Attachment
from spinta.manifests.components import Manifest
from spinta.backends.components import Backend
from spinta.types.datatype import File
from spinta.backends.fs.components import FileSystem


@commands.prepare.register(Context, FileSystem, Manifest)
def prepare(context: Context, backend: FileSystem, manifest: Manifest):
    pass


@commands.prepare.register(Context, FileSystem, File)
def prepare(context: Context, backend: FileSystem, dtype: File):
    pass


@commands.prepare_for_write.register(Context, File, Backend, Attachment)
def prepare_for_write(
    context: Context,
    dtype: File,
    backend: Backend,
    value: Attachment,
):
    return {
        '_id': value.filename,
        '_content_type': value.content_type,
    }



from typing import Any
from typing import Iterator
from typing import Optional

from starlette.requests import Request

from spinta.components import Context, Action, UrlParams, Node
from spinta import commands


def render(
    context: Context,
    request: Request,
    node: Node,
    params: UrlParams,
    data: Iterator[Any],
    *,
    action: Optional[Action] = None,
    status_code: Optional[int] = 200,
    headers: Optional[dict] = None,
):
    return commands.render(
        context, request, node, params.fmt,
        action=action,
        params=params,
        data=data,
        status_code=status_code,
        headers=headers,
    )



import ujson as json

from spinta.formats.components import Format


class JsonLines(Format):
    content_type = 'application/x-json-stream'
    accept_types = {
        'application/x-json-stream',
    }
    params = {}

    def __call__(self, data):
        for row in data:
            yield json.dumps(row, ensure_ascii=False) + '\n'



from starlette.requests import Request
from starlette.responses import FileResponse

from spinta import commands
from spinta.components import Action
from spinta.components import Context, Property, UrlParams
from spinta.backends.fs.components import FileSystem
from spinta.types.datatype import DataType, File
from spinta.exceptions import ItemDoesNotExist


@commands.getone.register(Context, Request, Property, File, FileSystem)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: File,
    backend: FileSystem,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, prop)

    accesslog = context.get('accesslog')
    accesslog.log(
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
    )

    data = getone(context, prop, prop.dtype, prop.model.backend, id_=params.pk)
    value = data[prop.name]
    filename = value['_id']
    if filename is None:
        raise ItemDoesNotExist(prop, id=params.pk)
    return FileResponse(
        str(dtype.backend.path / filename),
        media_type=value.get('_content_type'),
        headers={
            'Revision': data['_revision'],
            'Content-Disposition': (
                f'attachment; filename="{filename}"'
                if filename else
                'attachment'
            )
        },
    )


@commands.getone.register(Context, Property, DataType, FileSystem)
def getone(  # noqa
    context: Context,
    prop: Property,
    dtype: DataType,
    backend: FileSystem,
    *,
    id_: str,
):
    raise NotImplementedError


@commands.getone.register(Context, Property, File, FileSystem)
def getone(  # noqa
    context: Context,
    prop: Property,
    dtype: File,
    backend: FileSystem,
    *,
    id_: str,
):
    data = commands.getone(context, prop, prop.model.backend, id_=id_)
    if data is None:
        raise ItemDoesNotExist(prop, id=id_)
    data = (dtype.backend.path / data[prop.name]['_id']).read_bytes()
    return commands.cast_backend_to_python(context, prop, backend, data)



from pathlib import Path

from spinta.core.ufuncs import Bind
from spinta.core.ufuncs import Expr
from spinta.core.ufuncs import ufunc
from spinta.datasets.backends.sqldump.ufuncs.components import File
from spinta.datasets.backends.sqldump.ufuncs.components import PrepareFileResource
from spinta.exceptions import UnknownBind


@ufunc.resolver(PrepareFileResource, Expr)
def file(env: PrepareFileResource, expr: Expr) -> File:
    args, kwargs = expr.resolve(env)
    return env.call('file', *args, **kwargs)


@ufunc.resolver(PrepareFileResource, Bind)
def file(env: PrepareFileResource, path: Bind, **kwargs) -> File:
    if path.name == 'self':
        return env.call('file', env.path, **kwargs)
    else:
        raise UnknownBind(name=path.name)


@ufunc.resolver(PrepareFileResource, Path)
def file(env: PrepareFileResource, path: Path, *, encoding='utf-8') -> File:
    return File(path, encoding=encoding)



from starlette.requests import Request

from spinta import commands
from spinta.accesslog import AccessLog
from spinta.backends.helpers import get_select_prop_names
from spinta.backends.helpers import get_select_tree
from spinta.core.ufuncs import Expr
from spinta.renderer import render
from spinta.components import Context, Model, Property, Action, UrlParams
from spinta.types.datatype import DataType
from spinta.exceptions import NotFoundError, ItemDoesNotExist, UnavailableSubresource
from spinta.backends.postgresql.components import PostgreSQL
from spinta.utils.nestedstruct import flat_dicts_to_nested
from spinta.backends.postgresql.commands.query import PgQueryBuilder


@commands.getone.register(Context, Request, Model, PostgreSQL)
async def getone(
    context: Context,
    request: Request,
    model: Model,
    backend: PostgreSQL,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, model)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.log(
        model=model.model_type(),
        action=action.value,
        id_=params.pk,
    )

    data = getone(context, model, backend, id_=params.pk)
    select_tree = get_select_tree(context, action, params.select)
    prop_names = get_select_prop_names(
        context,
        model,
        model.properties,
        action,
        select_tree,
    )
    data = commands.prepare_data_for_response(
        context,
        model,
        params.fmt,
        data,
        action=Action.GETONE,
        select=select_tree,
        prop_names=prop_names,
    )
    return render(context, request, model, params, data, action=action)


@commands.getone.register(Context, Model, PostgreSQL)
def getone(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    id_: str,
):
    connection = context.get('transaction').connection
    table = backend.get_table(model)
    try:
        result = backend.get(connection, table, table.c._id == id_)
    except NotFoundError:
        raise ItemDoesNotExist(model, id=id_)
    data = flat_dicts_to_nested(dict(result))
    data['_type'] = model.model_type()
    return commands.cast_backend_to_python(context, model, backend, data)


@commands.getone.register(Context, Request, Property, DataType, PostgreSQL)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: DataType,
    backend: PostgreSQL,
    *,
    action: Action,
    params: UrlParams,
):
    raise UnavailableSubresource(prop=prop.name, prop_type=prop.dtype.name)


@commands.getall.register(Context, Model, PostgreSQL)
def getall(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    query: Expr = None,
):
    assert isinstance(query, (Expr, type(None))), query
    connection = context.get('transaction').connection

    builder = PgQueryBuilder(context)
    builder.update(model=model)
    table = backend.get_table(model)
    env = builder.init(backend, table)
    expr = env.resolve(query)
    where = env.execute(expr)
    qry = env.build(where)

    conn = connection.execution_options(stream_results=True)
    result = conn.execute(qry)

    for row in result:
        row = flat_dicts_to_nested(dict(row))
        row = {
            '_type': model.model_type(),
            **row,
        }
        row = commands.cast_backend_to_python(context, model, backend, row)
        yield row
